"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const categories_1 = __importDefault(require("./categories"));
const staticAnalysisCommon_1 = require("./staticAnalysisCommon");
const algorithmCategories_1 = __importDefault(require("./algorithmCategories"));
class lowLevelCalls {
    constructor() {
        this.llcNodes = [];
        this.name = `Low level calls: `;
        this.description = `Semantics maybe unclear`;
        this.category = categories_1.default.SECURITY;
        this.algorithm = algorithmCategories_1.default.EXACT;
    }
    visit(node) {
        if (staticAnalysisCommon_1.isLLCall(node)) {
            this.llcNodes.push({ node: node, type: staticAnalysisCommon_1.lowLevelCallTypes.CALL });
        }
        else if (staticAnalysisCommon_1.isLLDelegatecall(node)) {
            this.llcNodes.push({ node: node, type: staticAnalysisCommon_1.lowLevelCallTypes.DELEGATECALL });
        }
        else if (staticAnalysisCommon_1.isLLSend(node)) {
            this.llcNodes.push({ node: node, type: staticAnalysisCommon_1.lowLevelCallTypes.SEND });
        }
        else if (staticAnalysisCommon_1.isLLDelegatecall04(node)) {
            this.llcNodes.push({ node: node, type: staticAnalysisCommon_1.lowLevelCallTypes.DELEGATECALL });
        }
        else if (staticAnalysisCommon_1.isLLSend04(node)) {
            this.llcNodes.push({ node: node, type: staticAnalysisCommon_1.lowLevelCallTypes.SEND });
        }
        else if (staticAnalysisCommon_1.isLLCall04(node)) {
            this.llcNodes.push({ node: node, type: staticAnalysisCommon_1.lowLevelCallTypes.CALL });
        }
        else if (staticAnalysisCommon_1.isLLCallcode(node)) {
            this.llcNodes.push({ node: node, type: staticAnalysisCommon_1.lowLevelCallTypes.CALLCODE });
        }
    }
    report(compilationResults) {
        return this.llcNodes.map((item, i) => {
            let text = '';
            let morehref = '';
            switch (item.type) {
                case staticAnalysisCommon_1.lowLevelCallTypes.CALL:
                    text = `use of "call":  the use of low level "call" should be avoided whenever possible. 
                  It can lead to unexpected behavior if return value is not handled properly. 
                  Please use Direct Calls via specifying the called contract's interface.`;
                    morehref = 'http://solidity.readthedocs.io/en/develop/control-structures.html?#external-function-calls';
                    // http://solidity.readthedocs.io/en/develop/frequently-asked-questions.html?#why-is-the-low-level-function-call-less-favorable-than-instantiating-a-contract-with-a-variable-contractb-b-and-executing-its-functions-b-dosomething
                    break;
                case staticAnalysisCommon_1.lowLevelCallTypes.CALLCODE:
                    text = `use of "callcode":  the use of low level "callcode" should be avoided whenever possible. 
                  External code that is called can change the state of the calling contract and send ether from the caller's balance. 
                  If this is wanted behaviour use the Solidity library feature if possible.`;
                    morehref = 'http://solidity.readthedocs.io/en/develop/contracts.html#libraries';
                    break;
                case staticAnalysisCommon_1.lowLevelCallTypes.DELEGATECALL:
                    text = `use of "delegatecall": the use of low level "delegatecall" should be avoided whenever possible. 
                  External code that is called can change the state of the calling contract and send ether from the caller's balance. 
                  If this is wanted behaviour use the Solidity library feature if possible.`;
                    morehref = 'http://solidity.readthedocs.io/en/develop/contracts.html#libraries';
                    break;
                case staticAnalysisCommon_1.lowLevelCallTypes.SEND:
                    text = `use of "send": "send" does not throw an exception when not successful, make sure you deal with the failure case accordingly. 
                  Use "transfer" whenever failure of the ether transfer should rollback the whole transaction. 
                  Note: if you "send/transfer" ether to a contract the fallback function is called, the callees fallback function is very limited due to the limited amount of gas provided by "send/transfer". 
                  No state changes are possible but the callee can log the event or revert the transfer. "send/transfer" is syntactic sugar for a "call" to the fallback function with 2300 gas and a specified ether value.`;
                    morehref = 'http://solidity.readthedocs.io/en/develop/security-considerations.html#sending-and-receiving-ether';
                    break;
            }
            return { warning: text, more: morehref, location: item.node.src };
        });
    }
}
exports.default = lowLevelCalls;
//# sourceMappingURL=lowLevelCalls.js.map